package hikapro.com.backpack.model.dao;

import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteException;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.os.Handler;
import android.os.Looper;
import android.os.Message;
import android.os.Process;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import hikapro.com.backpack.App;
import hikapro.com.backpack.model.Api;
import hikapro.com.backpack.model.Model;
import hikapro.com.backpack.model.RestClient;
import hikapro.com.backpack.model.SetModel;
import hikapro.com.backpack.model.entities.Category;
import hikapro.com.backpack.model.entities.Item;
import hikapro.com.backpack.model.entities.Set;
import hikapro.com.backpack.model.entities.Timestamp;
import retrofit2.Call;
import retrofit2.Callback;
import retrofit2.Response;

/**
 * Created by tariel on 20/04/16.
 */
public class DAO {
    //region Constants
    private static final int CORE_POOL_SIZE = 1;
    private static final int CORE_MAX_POOL_SIZE = 1;

    private static final int KEEP_ALIVE_TIME = 1;
    private static final TimeUnit KEEP_ALIVE_TIME_UNIT = TimeUnit.SECONDS;

    private static final int MY_LIST_ID = 15;
    //endregion

    private static DAO instance;

    private final ThreadPoolExecutor threadPool;
    private final Handler   handler;
    private SetModel        setModel;
    private final Api       api;
    private Context         context;
    private DbHelper        helper;
    private Map<String, Model.Base> observers;

    private DAO() {
        this.context    = App.getAppContext();
        this.helper     = new DbHelper(this.context);
        this.api        = RestClient.getApi();
        final BlockingQueue<Runnable> taskQueue = new LinkedBlockingQueue<>();
        this.threadPool = initPool(taskQueue);
        this.observers  = Collections.synchronizedMap(new HashMap<String, Model.Base>());
        this.handler    = initHandler();
    }

    static {
        instance = new DAO();
    }

    public static DAO getInstance() {
        return instance;
    }

    public void registerObserver(Model.Base o) {
        observers.put(o.getClass().getName(), o);
    }

    private ThreadPoolExecutor initPool (BlockingQueue<Runnable> taskQueue) {
        ThreadPoolExecutor ret = new ThreadPoolExecutor(
                CORE_POOL_SIZE
                ,CORE_MAX_POOL_SIZE
                ,KEEP_ALIVE_TIME
                ,KEEP_ALIVE_TIME_UNIT
                ,taskQueue
        );
        return ret;
    }

    private Handler initHandler() {
        Handler ret = new Handler(Looper.getMainLooper()) {
            @Override
            public void handleMessage(Message msg) {
                for (Map.Entry<String, Model.Base> entry : observers.entrySet()) {
                    entry.getValue().onEvent(msg);
                }
            }
        };
        return ret;
    }

    public void executeCommand(Message command) {

        SetTask     setTask;
        ItemTask    itemTask;
        ImageProviderTask imageProviderTask;

        if (command != null) {
            switch (command.what) {

                case Command.SET_GET_ALL :
                    setTask = new SetTask(Command.SET_GET_ALL,
                            Process.THREAD_PRIORITY_MORE_FAVORABLE);
                    threadPool.execute(setTask);
                    break;

                case Command.SET_REORDER :
                    setTask = new SetTask(Command.SET_REORDER,
                            Process.THREAD_PRIORITY_BACKGROUND);
                    setTask.setsToUpdate = (List<Set>) command.obj;
                    threadPool.execute(setTask);
                    break;

                case Command.SET_GET_ITEMS :
                    setTask = new SetTask(Command.SET_GET_ITEMS,
                            Process.THREAD_PRIORITY_MORE_FAVORABLE);
                    setTask.setId = command.arg1;
                    threadPool.execute(setTask);
                    break;

                case Command.SET_GET_PACKED_ITEMS :
                    setTask = new SetTask(Command.SET_GET_PACKED_ITEMS,
                            Process.THREAD_PRIORITY_MORE_FAVORABLE);
                    setTask.setId = command.arg1;
                    threadPool.execute(setTask);
                    break;

                case Command.SET_UNPACK_ITEMS :
                    setTask = new SetTask(Command.SET_UNPACK_ITEMS,
                            Process.THREAD_PRIORITY_DEFAULT);
                    setTask.setId = command.arg1;
                    threadPool.execute(setTask);

                case Command.ITEM_GET_CATEGORIES :
                    itemTask = new ItemTask(Command.ITEM_GET_CATEGORIES,
                            Process.THREAD_PRIORITY_MORE_FAVORABLE);
                    threadPool.execute(itemTask);
                    break;

                case Command.ITEM_INSERT :
                    itemTask = new ItemTask(Command.ITEM_INSERT,
                            Process.THREAD_PRIORITY_BACKGROUND);
                    itemTask.item   = (Item) command.obj;
                    itemTask.setId  = command.arg1;
                    threadPool.execute(itemTask);
                    break;

                case Command.ITEM_DELETE_FROM_SET :
                    itemTask = new ItemTask(Command.ITEM_DELETE_FROM_SET,
                            Process.THREAD_PRIORITY_BACKGROUND);
                    itemTask.setId  = command.arg1;
                    itemTask.itemId = command.arg2;
                    threadPool.execute(itemTask);
                    break;

                case Command.ITEM_PACK :
                    itemTask = new ItemTask(Command.ITEM_PACK,
                            Process.THREAD_PRIORITY_BACKGROUND);
                    itemTask.setId  = command.arg1;
                    itemTask.itemId = command.arg2;
                    threadPool.execute(itemTask);
                    break;

                case Command.ITEM_UNPACK :
                    itemTask = new ItemTask(Command.ITEM_UNPACK,
                            Process.THREAD_PRIORITY_BACKGROUND);
                    itemTask.setId  = command.arg1;
                    itemTask.itemId = command.arg2;
                    threadPool.execute(itemTask);
                    break;

                case Command.ITEM_GET_IMAGE :
                    Item item = findItem(command.arg1);
                    if (item != null) {
                        imageProviderTask = new ImageProviderTask(Command.ITEM_GET_IMAGE,
                                Process.THREAD_PRIORITY_DEFAULT);
                        imageProviderTask.item = item;
                        threadPool.execute(imageProviderTask);
                    }
                    break;

                case Command.MY_LIST_ITEM_ADD :
                    break;

                case Command.MY_LIST_ITEM_DELETE :
                    break;

                case Command.MY_LIST_POST :
                    break;

                case Command.MY_LIST_CLEAR :
                    break;

                case Command.SYNC :
                    threadPool.execute(new SyncTask(Command.SYNC,
                            Process.THREAD_PRIORITY_BACKGROUND));
                    break;

                case Command.SYNC_IF_NOT_EXISTS :
                    threadPool.execute(new SyncTask(Command.SYNC_IF_NOT_EXISTS,
                            Process.THREAD_PRIORITY_MORE_FAVORABLE));
                    break;
            }
        }
    }

    ///////////////////////  DATABASE  /////////////////////

    //region Database
    private SQLiteDatabase getReadDB(){
        return helper.getReadableDatabase();
    }

    private SQLiteDatabase getWriteDB(){
        return helper.getWritableDatabase();
    }

    // inserts
    private void insertTimestamp(Timestamp timestamp) {
        if (timestamp != null && timestamp.timestamp > 0) {
            ContentValues values;
            SQLiteDatabase db = getWriteDB();
            try {
                db.beginTransaction();
                values = Db.LogTable.toContentValues(timestamp);
                db.insert(Db.LogTable.TABLE_NAME, null, values);
                db.setTransactionSuccessful();
            } finally {
                db.endTransaction();
                db.close();
            }
        }
    }
    private void insertItems(List<Item> items) {
        if (items != null && !items.isEmpty()) {
            ContentValues values;
            SQLiteDatabase db = getWriteDB();
            try {
                db.beginTransaction();
                for (Item item : items) {
                    values = Db.ItemsTable.toContentValues(item);
                    db.insert(Db.ItemsTable.TABLE_NAME, null, values);
                }
                db.setTransactionSuccessful();
            } finally {
                db.endTransaction();
                db.close();
            }
        }
    }
    private void insertCategories(List<Category> categories) {

        if (categories != null && !categories.isEmpty()) {
            ContentValues values;
            SQLiteDatabase db = getWriteDB();
            try {
                db.beginTransaction();
                for (Category category : categories) {
                    values = Db.CategoriesTable.toContentValues(category);
                    db.insert(Db.CategoriesTable.TABLE_NAME, null, values);
                }
                db.setTransactionSuccessful();
            } finally {
                db.endTransaction();
                db.close();
            }
        }
    }
    private void insertSets(List<Set> sets) {

        if (sets != null && !sets.isEmpty()) {
            ContentValues values;
            int i = 0;
            SQLiteDatabase db = getWriteDB();
            try {
                db.beginTransaction();
                for (Set set : sets) {
                    values = Db.SetsTable.toContentValues(set, i);
                    db.insert(Db.SetsTable.TABLE_NAME, null, values);
                    insertSetItems(set, db);
                    ++i;
                }
                db.setTransactionSuccessful();
            } finally {
                db.endTransaction();
                db.close();
            }
        }
    }
    private void insertSetItems(Set set, SQLiteDatabase db) {
        if (set != null && db != null) {
            if (!set.getItems().isEmpty()) {
                ContentValues values;
                int setId = set.getId();
                try {
                    db.beginTransaction();
                    for (Integer itemid : set.getItems()) {
                        values = Db.SetItemsTable.toContentValues(setId, itemid);
                        db.insert(Db.SetItemsTable.TABLE_NAME, null, values);
                    }
                    db.setTransactionSuccessful();
                } finally {
                    db.endTransaction();
                }
            }
        }
    }
    private void insertSetItem(int setId, int itemId) {
        ContentValues values;
        SQLiteDatabase db = getWriteDB();
        try {
            db.beginTransaction();
            values = Db.SetItemsTable.toContentValues(setId, itemId);
            db.insert(Db.SetItemsTable.TABLE_NAME, null, values);
            db.setTransactionSuccessful();
        } finally {
            db.endTransaction();
            db.close();
        }
    }
    // reads
    private boolean LogExist() {
        boolean ret;
        SQLiteDatabase db = getReadDB();
        Cursor cursor = db.query(Db.LogTable.TABLE_NAME,
                new String[] {Db.LogTable.COLUMN_ID}, null, null, null, null, null, "1");
        ret = cursor.moveToNext();
        cursor.close();
        db.close();
        return ret;
    }

    private Item findItem(int id) {
        Item ret = null;
        Cursor cursor = null;
        SQLiteDatabase db = null;
        try {
            db = getReadDB();
            String query = String.format("SELECT * FROM %s a WHERE a.%s = %d LIMIT 1",
                    Db.ItemsTable.TABLE_NAME, Db.ItemsTable.COLUMN_ID, id);
            cursor = db.rawQuery(query, null);
            if (cursor.moveToNext()) {
                ret = Db.ItemsTable.parseCursor(cursor);
            }
        } catch (Exception e) {
        //TODO write to log here
        } finally {
            if (cursor != null)
                cursor.close();
            if (db != null)
                db.close();
        }
        return ret;
    }

    private List<Item> readItems(int setId, boolean packed) {
        List<Item> ret = new ArrayList<>(256);
        Cursor cursor = null;
        SQLiteDatabase db = null;
        Item item;
        String query = String.format(
                "SELECT a.%s, a.%s, a.%s, a.%s, a.%s, a.%s, a.%s, a.%s, a.%s FROM %s a INNER JOIN %s b ON a.%s = b.%s WHERE b.%s = ? AND b.%s <> 1 AND b.%s %s 1",

                Db.ItemsTable.COLUMN_ID,
                Db.ItemsTable.COLUMN_NAME,
                Db.ItemsTable.COLUMN_CATEGORY,
                Db.ItemsTable.COLUMN_DESCRIPTION,
                Db.ItemsTable.COLUMN_BUY_URLS,
                Db.ItemsTable.COLUMN_PHOTO_URL,
                Db.ItemsTable.COLUMN_PHOTO_THUMB_URL,
                Db.ItemsTable.COLUMN_PHOTO_LOCAL,
                Db.ItemsTable.COLUMN_PHOTO_THUMB_LOCAL,

                Db.ItemsTable.TABLE_NAME,
                Db.SetItemsTable.TABLE_NAME,
                Db.ItemsTable.COLUMN_ID,
                Db.SetItemsTable.COLUMN_ITEM,
                Db.SetItemsTable.COLUMN_SET,
                Db.SetItemsTable.COLUMN_DELETED,
                Db.SetItemsTable.COLUMN_PACKED,
                packed ? "=" : "<>");
        try {
            db = getReadDB();
            cursor = db.rawQuery(query, new String[]{String.valueOf(setId)});
            while (cursor.moveToNext()) {
                item = Db.ItemsTable.parseCursor(cursor);
                ret.add(item);
            }
        } catch (SQLiteException e) {
            //TODO write to log here

        } catch (Exception e) {
            //TODO write to log here
        } finally {
            if (cursor != null)
                cursor.close();
            if (db != null)
                db.close();
        }
        return ret;
    }
    private Hashtable<Integer, Category> readCategories() {
        Hashtable<Integer, Category> ret = new Hashtable<>(20, 0.9f);
        Cursor cursor = null;
        SQLiteDatabase db = null;
        Category category;
        try {
            db = getReadDB();
            cursor = db.query(Db.CategoriesTable.TABLE_NAME,
                    new String[]{Db.CategoriesTable.COLUMN_ID,
                            Db.CategoriesTable.COLUMN_NAME},
                    null,null,null,null,null);
            while (cursor.moveToNext()) {
                category = Db.CategoriesTable.parseCursor(cursor);
                ret.put(category.getId(), category);
            }
        } catch (SQLiteException e) {
            //TODO write to log here

        } catch (Exception e) {
            //TODO write to log here
        } finally {
            if (cursor != null)
                cursor.close();
            if (db != null)
                db.close();
        }
        return ret;
    }

    private Set readSet(SQLiteDatabase db, int setId) {
        Set ret = null;
        Cursor cursor;
        String query = String.format("SELECT * FROM %s a WHERE a.%s = %d LIMIT 1",
                Db.SetsTable.TABLE_NAME, Db.SetsTable.COLUMN_ID, setId);
        cursor = db.rawQuery(query, null);
        if (cursor.moveToNext())
            ret = Db.SetsTable.parseCursor(cursor);
        if (cursor != null)
            cursor.close();
        return ret;
    }

    private List<Set> readSets() {
        List<Set> ret = new ArrayList<>(12);
        Cursor cursor = null;
        SQLiteDatabase db = null;
        Set set;
        try {
            db = getReadDB();
            String query = String.format("SELECT * FROM %s", Db.SetsTable.TABLE_NAME);
            cursor = db.rawQuery(query, null);
            while (cursor.moveToNext()) {
                set = Db.SetsTable.parseCursor(cursor);
                ret.add(set);
            }
            Collections.sort(ret);
        } catch (SQLiteException e) {
            //TODO write to log here

        } catch (Exception e) {
            //TODO write to log here
        } finally {
            if (cursor != null)
                cursor.close();
            if (db != null)
                db.close();
        }
        return ret;
    }
    // updates
    private int updateItemLocalPic(int id, String path) {
        int ret = 0;
        SQLiteDatabase db = null;
        ContentValues values;
        try {
            db = getWriteDB();
            db.beginTransaction();
            values = new ContentValues();
            values.put(Db.ItemsTable.COLUMN_PHOTO_LOCAL, path);
            ret = db.update(Db.ItemsTable.TABLE_NAME, values, "_id = ?",
                    new String[]{String.valueOf(id)});
            db.setTransactionSuccessful();
        } catch (SQLiteException e) {
            //TODO write to log here

        } catch (Exception e) {
            //TODO write to log here
        } finally {
            if (db != null) {
                db.endTransaction();
                db.close();
            }
        }
        return ret;
    }

    private int updateSetsOrder(List<Set> reorderedSet) {
        int ret = 0;
        SQLiteDatabase db = null;
        ContentValues values;
        try {

            db = getWriteDB();
            db.beginTransaction();
            for (Set set : reorderedSet) {
                values = new ContentValues();
                values.put(Db.SetsTable.COLUMN_LINE_NUMBER, set.getLineNumber());
                ret += db.update(Db.SetsTable.TABLE_NAME, values, "_id = ?",
                        new String[]{String.valueOf(set.getId())});
            }
            db.setTransactionSuccessful();
        } catch (SQLiteException e) {
            //TODO write to log here

        } catch (Exception e) {
            //TODO write to log here
        } finally {
            if (db != null) {
                db.endTransaction();
                db.close();
            }
        }
        return ret;
    }
    private int updateSetItemDeleted(int setId, int itemId, boolean del) {
        int ret = 0;
        SQLiteDatabase db = null;
        ContentValues values;
        try {
            db = getWriteDB();
            db.beginTransaction();
            values = new ContentValues();
            values.put(Db.SetItemsTable.COLUMN_DELETED, del);
            ret = db.update(Db.SetItemsTable.TABLE_NAME, values, String.format("%s = ? AND %s = ?",
                    Db.SetItemsTable.COLUMN_SET, Db.SetItemsTable.COLUMN_ITEM ),
                    new String[]{String.valueOf(setId), String.valueOf(itemId)});
            db.setTransactionSuccessful();
        } catch (SQLiteException e) {
            //TODO write to log here

        } catch (Exception e) {
            //TODO write to log here
        } finally {
            if (db != null) {
                db.endTransaction();
                db.close();
            }
        }
        return ret;
    }

    private int updateSetItemsPacked(int setId, boolean pack) {
        int  ret = 0;
        SQLiteDatabase db = null;
        ContentValues values;
        Set set;
        try {
            db = getWriteDB();
            db.beginTransaction();

            /*
            String query = String.format("UPDATE %s SET %s = %d where %s = %d",
                    Db.SetItemsTable.TABLE_NAME,
                    Db.SetItemsTable.COLUMN_PACKED,
                    pack,
                    Db.SetItemsTable.COLUMN_SET,
                    setId);
            Cursor cursor = db.rawQuery(query, null);*/

            values = new ContentValues();
            values.put(Db.SetItemsTable.COLUMN_PACKED, pack);
            ret = db.update(Db.SetItemsTable.TABLE_NAME, values, String.format("%s = ?",
                    Db.SetItemsTable.COLUMN_SET),
                    new String[]{String.valueOf(setId)});

            set = readSet(db, setId);
            if (set != null) {
                values = new ContentValues();
                values.put(Db.SetsTable.COLUMN_PACKED_QTY, pack ? set.getActiveQty() : 0);
                ret += db.update(Db.SetsTable.TABLE_NAME, values, String.format("%s = ?",
                        Db.SetsTable.COLUMN_ID),
                        new String[]{String.valueOf(setId)});
            }
            db.setTransactionSuccessful();
        } catch (SQLiteException e) {
            //TODO write to log here

        } catch (Exception e) {
            //TODO write to log here
        } finally {
            if (db != null) {
                db.endTransaction();
                db.close();
            }
        }
        return ret;
    }

    private int updateSetItemPacked(int setId, int itemId, boolean pack) {
        int ret = 0;
        SQLiteDatabase db = null;
        ContentValues values;
        Set set;
        try {
            db = getWriteDB();
            db.beginTransaction();

            values = new ContentValues();
            values.put(Db.SetItemsTable.COLUMN_PACKED, pack);
            ret = db.update(Db.SetItemsTable.TABLE_NAME, values, String.format("%s = ? AND %s = ?",
                    Db.SetItemsTable.COLUMN_SET, Db.SetItemsTable.COLUMN_ITEM ),
                    new String[]{String.valueOf(setId), String.valueOf(itemId)});

            set = readSet(db, setId);
            if (set != null) {
                values = new ContentValues();
                values.put(Db.SetsTable.COLUMN_PACKED_QTY, pack ? set.getPackedQty() + 1 : set.getPackedQty() - 1);
                ret += db.update(Db.SetsTable.TABLE_NAME, values, String.format("%s = ?",
                        Db.SetsTable.COLUMN_ID),
                        new String[]{String.valueOf(setId)});
            }
            db.setTransactionSuccessful();
        } catch (SQLiteException e) {
            //TODO write to log here

        } catch (Exception e) {
            //TODO write to log here
        } finally {
            if (db != null) {
                db.endTransaction();
                db.close();
            }
        }
        return ret;
    }
    //endregion

    ///////////////////////  TASK CLASSES //////////////////

    //region Task classes

    // MY LIST CLASS
    private class MyListTask implements Runnable {
        int currentCommand;
        int priority;


        public MyListTask(int command, int priority) {
            this.currentCommand = command;
            this.priority       = priority;
        }

        @Override
        public void run() {
            android.os.Process.setThreadPriority(priority);
            Message message = Message.obtain();
            switch (currentCommand) {

            }
            handler.sendMessage(message);
        }
    }
    // ITEM CLASS
    private class ItemTask implements Runnable {
        int currentCommand;
        int priority;
        int setId;
        int itemId;
        Item item;

        public ItemTask(int command, int priority) {
            this.currentCommand = command;
            this.setId          = -1;
            this.itemId         = -1;
            this.priority       = priority;
        }

        @Override
        public void run() {
            android.os.Process.setThreadPriority(priority);
            Message message = Message.obtain();
            switch (currentCommand) {

                case Command.ITEM_DELETE_FROM_SET :
                    message.arg1 = updateSetItemDeleted(setId, itemId, true);
                    if (message.arg1 > 0)
                        message.what = Event.ITEM_FROM_SET_DELETED;
                    else
                        message.what = Event.ITEM_FROM_SET_ERROR;
                    break;

                case Command.ITEM_INSERT :
                    List<Item> items = new ArrayList<>();
                    items.add(item);
                    if (items.isEmpty())
                        message.what = Event.ITEM_INSERT_ERROR;
                    else {
                        insertItems(items);
                        insertSetItem(setId, item.getId());
                        message.what = Event.ITEM_INSERTED;
                        message.arg1 = setId;
                        message.arg2 = item.getId();
                    }
                    break;

                case Command.ITEM_PACK :
                    message.arg1 = updateSetItemPacked(setId, itemId, true);
                    if (message.arg1 > 0)
                        message.what = Event.ITEM_PACKED;
                    else
                        message.what = Event.ITEM_PACK_ERROR;
                    break;

                case Command.ITEM_UNPACK :
                    message.arg1 = updateSetItemPacked(setId, itemId, false);
                    if (message.arg1 > 0)
                        message.what = Event.ITEM_UNPACKED;
                    else
                        message.what = Event.ITEM_UNPACK_ERROR;
                    break;

                case Command.ITEM_GET_CATEGORIES :
                    Hashtable<Integer, Category> res = readCategories();
                    if (res.isEmpty())
                        message.what = Event.ITEM_CATEGORY_LOAD_ERROR;
                    else {
                        message.what = Event.ITEM_CATEGORY_LOAD_COMPLETED;
                        message.obj = res;
                    }
                    break;
            }
            handler.sendMessage(message);
        }
    }
    // SET CLASS
    private class SetTask implements Runnable {
        int currentCommand;
        int priority;
        int setId;
        List<Set> setsToUpdate;
        List<Item> items;

        public SetTask(int command, int priority) {
            this.currentCommand = command;
            this.setId          = -1;
            this.setsToUpdate   = new ArrayList<>();
            this.priority       = priority;
        }

        @Override
        public void run() {
            android.os.Process.setThreadPriority(priority);
            Message message = Message.obtain();
            switch (currentCommand) {

                case Command.SET_GET_ALL :
                    List<Set> sets = readSets();
                    if (sets.isEmpty())
                        message.what = Event.SET_LOAD_ERROR;
                    else {
                        message.what = Event.SET_LOAD_COMPLETED;
                        message.obj = sets;
                    }
                    break;

                case Command.SET_GET_ITEMS :
                    items = readItems(setId, false);
                    if (items.isEmpty())
                        message.what = Event.SET_ITEMS_LOAD_ERROR;
                    else {
                        Collections.sort(items);
                        message.what = Event.SET_ITEMS_LOAD_COMPLETED;
                        message.obj = items;
                        message.arg1 = setId;
                    }
                    break;

                case Command.SET_REORDER :
                    message.arg1 = updateSetsOrder(setsToUpdate);
                    if (message.arg1 > 0)
                        message.what = Event.SET_REORDER_COMPLETED;
                    else
                        message.what = Event.SET_REORDER_ERROR;
                    break;

                case Command.SET_GET_PACKED_ITEMS :
                    items = readItems(setId, true);
                    if (items.isEmpty())
                        message.what = Event.SET_PACKED_LOAD_ERROR;
                    else {
                        Collections.sort(items);
                        message.what = Event.SET_PACKED_LOAD_COMPLETED;
                        message.obj = items;
                        message.arg1 = setId;
                    }
                    break;

                case Command.SET_UNPACK_ITEMS :
                    message.arg1 = updateSetItemsPacked(setId, false);
                    message.arg2 = setId;
                    if (message.arg1 > 0)
                        message.what = Event.SET_UNPACK_COMPLETED;
                    else
                        message.what = Event.SET_UNPACK_ERROR;
                    break;
            }
            handler.sendMessage(message);
        }
    }
    // SYNC CLASS
    private class SyncTask implements Runnable {
        int currentCommand;
        int priority;
        int statusCode;

        public SyncTask(int command, int priority) {
            this.currentCommand = command;
            this.priority       = priority;
        }

        @Override
        public void run() {
            android.os.Process.setThreadPriority(priority);
            Message message = Message.obtain();

            switch (currentCommand) {
                case Command.SYNC:
                    try {
                        Call<List<Set>> call = api.getSets();
                        call.enqueue(new Callback<List<Set>>() {
                            @Override
                            public void onResponse(Call<List<hikapro.com.backpack.model.entities.Set>> call, Response<List<Set>> response) {
                                statusCode = response.code();
                                // TODO
                                // check if first time
                                // if not check for updates else
                                // insert into database here
                                insertSets(response.body());
                            }
                            @Override
                            public void onFailure(Call<List<hikapro.com.backpack.model.entities.Set>> call, Throwable t) {
                            }
                        });
                        message.what = Event.SYNC_COMPLETED;

                    } catch (Exception e) {
                        message.what = Event.SYNC_FAILED;
                    }
                    finally {
                        message.arg1 = statusCode;
                        handler.sendMessage(message);
                    }
                    break;

                case Command.SYNC_IF_NOT_EXISTS:
                    if (LogExist()) {
                        message.what = Event.SYNC_COMPLETED;
                    } else {

                        try {
                            Response<List<Set>> response0 = api.getSets().execute();
                            insertSets(response0.body());
                            statusCode = response0.code();
                            Response<List<Category>> response1 = api.getItemCategories().execute();
                            insertCategories(response1.body());
                            statusCode = response1.code();
                            Response<List<Item>> response2 = api.getItems().execute();
                            insertItems(response2.body());
                            statusCode = response2.code();
                            Response<Timestamp> response3 = api.getTimestamp().execute();
                            insertTimestamp(response3.body());
                            statusCode = response3.code();
                            message.what = Event.SYNC_COMPLETED;
                        } catch (IOException e ){
                            message.what = Event.SYNC_FAILED;
                        } finally {
                            message.arg1 = statusCode;
                            handler.sendMessage(message);
                        }
                    }
                    break;
            }
        }
    }
    // IMAGE PROVIDER CLASS
    private class ImageProviderTask implements Runnable {
        int currentCommand;
        int priority;
        Item item;


        public ImageProviderTask(int command, int priority) {
            this.currentCommand = command;
            this.priority       = priority;
        }

        @Override
        public void run() {
            android.os.Process.setThreadPriority(priority);
            Message message = Message.obtain();
            switch (currentCommand) {

                case Command.ITEM_GET_IMAGE :
                    try {
                        Bitmap bitmap = loadItemImage(item);
                        if (bitmap != null) {
                            message.what = Event.ITEM_IMAGE_LOAD_COMPLETED;
                            message.obj = bitmap;
                        }
                        else
                            message.what = Event.ITEM_IMAGE_LOAD_ERROR;
                    } catch (Exception e) {
                        message.what = Event.ITEM_IMAGE_LOAD_ERROR;
                    }
                    break;
            }
            handler.sendMessage(message);
        }
    }
    //endregion

    private Bitmap loadItemImage(Item item){
        ImageDownloadHelper downloadHelper = new ImageDownloadHelper();
        Bitmap bitmap   = null;
        String filename = null;

        if (item != null) {
            bitmap = BitmapFactory.decodeFile(item.getPhotoLocal());
            // cannot retrieve, download and save then
            if (bitmap == null) { // return it
                bitmap = downloadHelper.loadImage(item.getPhotoUrl());
                if (bitmap != null) {
                    if (downloadHelper.isExternalStorageWritable()) {
                        filename = downloadHelper.saveImageExternal(
                                downloadHelper.generateFileName(item.getPhotoUrl()), bitmap);
                    } else {
                        filename = downloadHelper.saveImageInternal(
                                downloadHelper.generateFileName(item.getPhotoUrl()), bitmap);
                    }
                    updateItemLocalPic(item.getId(), filename);
                }
            }
        }
        return bitmap;
    }


}
